import logging
import datetime
from collections import OrderedDict
from PyQt4 import Qt
from PyQt4 import QtCore


class _DownloadStat:
    def __init__(self):
        self.done_bytes = 0
        self.dt = None


class DownloadItemState:
    QUEUED = 0
    RUNNING = 1
    PAUSED = 2
    CANCELLED = 3
    ERROR = 4
    DONE = 5


class _DownloadItem:
    def __init__(self, work):
        self._work = work
        self._dl_progress = None
        self._total_segments = None
        self._filename = None
        self._added_dt = datetime.datetime.now()
        self._started_dt = None
        self._end_elapsed = None
        self._last_dl_stat = None
        self._avg_speed = 0
        self._error = None
        self._state = DownloadItemState.QUEUED

    def set_error(self, ex):
        self._error = ex

    def get_error(self):
        return self._error

    def get_state(self):
        return self._state

    def set_state(self, state):
        if state == DownloadItemState.RUNNING:
            self._started_dt = datetime.datetime.now()
        elif state in [
            DownloadItemState.DONE,
            DownloadItemState.CANCELLED,
            DownloadItemState.ERROR
        ]:
            self._end_elapsed = self.get_elapsed()
            self._avg_speed = 0

        self._state = state

    def get_dl_progress(self):
        return self._dl_progress

    def get_avg_download_speed(self):
        return self._avg_speed

    def _compute_avg_speed(self, dt):
        done_bytes = self.get_dl_progress().get_done_bytes()
        now = dt

        if self._last_dl_stat is None:
            self._last_dl_stat = _DownloadStat()
            self._last_dl_stat.done_bytes = done_bytes
            self._last_dl_stat.dt = now
            return

        time_delta = now - self._last_dl_stat.dt
        time_delta = time_delta.total_seconds()
        bytes_delta = done_bytes - self._last_dl_stat.done_bytes
        last_speed = bytes_delta / time_delta
        self._avg_speed = 0.2 * last_speed + 0.8 * self._avg_speed

        self._last_dl_stat.done_bytes = done_bytes
        self._last_dl_stat.dt = now

    def set_dl_progress(self, dl_progress, dt):
        self._dl_progress = dl_progress

        if self.get_state() == DownloadItemState.RUNNING:
            self._compute_avg_speed(dt)

    def get_work(self):
        return self._work

    def set_work(self, work):
        self._work = work

    def get_total_segments(self):
        return self._total_segments

    def set_total_segments(self, total_segments):
        self._total_segments = total_segments

    def get_filename(self):
        return self._filename

    def set_filename(self, filename):
        self._filename = filename

    def get_progress_percent(self):
        is_init = (self.get_state() == DownloadItemState.QUEUED)
        if is_init or self.get_dl_progress() is None:
            return 0
        if self.get_state() == DownloadItemState.DONE:
            return 100

        num = self.get_dl_progress().get_done_segments()
        denom = self.get_total_segments()

        return round(num / denom * 100)

    def get_added_dt(self):
        return self._added_dt

    def get_started_dt(self):
        return self._started_dt

    def get_elapsed(self):
        if self.get_state() == DownloadItemState.QUEUED:
            return datetime.timedelta()
        elif self._end_elapsed is not None:
            return self._end_elapsed

        return datetime.datetime.now() - self.get_started_dt()

    def get_estimated_size(self):
        if self.get_state() == DownloadItemState.DONE:
            return self.get_dl_progress().get_done_bytes()

        if self.get_dl_progress() is None:
            return None

        done_segments_bytes = self.get_dl_progress().get_done_segments_bytes()
        done_segments = self.get_dl_progress().get_done_segments()
        total_segments = self.get_total_segments()

        if done_segments == 0 or done_segments_bytes == 0:
            return None

        estimated_size = total_segments / done_segments * done_segments_bytes

        return estimated_size


class QDownloadsTableModel(Qt.QAbstractTableModel):
    _HEADER = [
        'Emission',
        'Season/ep.',
        'Episode',
        'Filename',
        'Sections',
        'Downloaded',
        'Estimated size',
        'Added',
        'Elapsed',
        'Speed',
        'Progress',
        'Status',
        'Quality',
    ]
    _status_msg_handlers = {
        DownloadItemState.QUEUED: lambda i: 'Queued',
        DownloadItemState.RUNNING: lambda i: 'Running',
        DownloadItemState.PAUSED: lambda i: 'Paused',
        DownloadItemState.CANCELLED: lambda i: 'Cancelled',
        DownloadItemState.ERROR: lambda i: 'Error: {}'.format(i.get_error()),
        DownloadItemState.DONE: lambda i: 'Done'
    }
    download_finished = QtCore.pyqtSignal(object)
    download_cancelled = QtCore.pyqtSignal(object)

    def __init__(self, download_manager, parent=None):
        super().__init__(parent)

        self._download_manager = download_manager
        self._download_list = OrderedDict()

        self._delayed_update_calls = []

        self._setup_signals()
        self._setup_timer()

    def get_progress_col(self):
        return 10

    def get_download_item_at_row(self, row):
        return list(self._download_list.values())[row]

    def download_item_exists(self, episode_id, quality):
        key = (episode_id, quality)
        return key in self._download_list

    def cancel_download_at_row(self, row):
        # Get download item
        dl_item = self.get_download_item_at_row(row)

        # Ask download manager to cancel its work
        self._download_manager.cancel_work(dl_item.get_work())

    def remove_episode_id_item(self, episode_id, quality):
        key = (episode_id, quality)
        if key not in self._download_list:
            return

        row = list(self._download_list.keys()).index(key)
        self.beginRemoveRows(Qt.QModelIndex(), row, row)
        del self._download_list[key]
        self.endRemoveRows()

    def remove_item_at_row(self, row):
        self.beginRemoveRows(Qt.QModelIndex(), row, row)
        key = list(self._download_list.keys())[row]
        del self._download_list[key]
        self.endRemoveRows()

    def _setup_timer(self):
        self._refresh_timer = Qt.QTimer(self)
        self._refresh_timer.timeout.connect(self._on_timer_timeout)
        self._refresh_timer.setInterval(500)
        self._refresh_timer.start()

    def _setup_signals(self):
        dlman = self._download_manager

        dlman.download_created.connect(self._on_download_created_delayed)
        dlman.download_started.connect(self._on_download_started_delayed)
        dlman.download_progress.connect(self._on_download_progress_delayed)
        dlman.download_finished.connect(self._on_download_finished_delayed)
        dlman.download_error.connect(self._on_download_error_delayed)
        dlman.download_cancelled.connect(self._on_download_cancelled_delayed)

    def _on_download_created_delayed(self, work):
        self._delayed_update_calls.append((self._on_download_created, [work]))

    def _on_download_started_delayed(self, work, dl_progress, filename,
                                     total_segments):
        now = datetime.datetime.now()
        self._delayed_update_calls.append(
            (self._on_download_started, [work, dl_progress, filename,
                                         total_segments, now]))

    def _on_download_progress_delayed(self, work, dl_progress):
        now = datetime.datetime.now()
        self._delayed_update_calls.append((self._on_download_progress,
                                          [work, dl_progress, now]))

    def _on_download_finished_delayed(self, work):
        self._delayed_update_calls.append((self._on_download_finished, [work]))

    def _on_download_error_delayed(self, work, ex):
        self._delayed_update_calls.append((self._on_download_error,
                                           [work, ex]))

    def _on_download_cancelled_delayed(self, work):
        self._delayed_update_calls.append((self._on_download_cancelled,
                                           [work]))

    def _on_download_created(self, work):
        episode_id = work.get_episode().get_id()
        quality = work.quality
        key = (episode_id, quality)

        if key in self._download_list:
            msg = 'Episode {} already in download list'.format(episode_id)
            logging.warning(msg)
            return

        new_position = len(self._download_list)
        self.beginInsertRows(Qt.QModelIndex(), new_position, new_position)
        self._download_list[key] = _DownloadItem(work)
        self.endInsertRows()

    def _get_download_item(self, episode, quality):
        key = (episode.get_id(), quality)
        return self._download_list[key]

    def _on_download_started(self, work, dl_progress, filename,
                             total_segments, now):
        episode = work.get_episode()
        quality = work.quality

        item = self._get_download_item(episode, quality)

        item.set_dl_progress(dl_progress, now)
        item.set_total_segments(total_segments)
        item.set_filename(filename)
        item.set_state(DownloadItemState.RUNNING)

    def _on_download_progress(self, work, dl_progress, now):
        episode = work.get_episode()
        quality = work.quality

        item = self._get_download_item(episode, quality)

        item.set_dl_progress(dl_progress, now)

    def _on_download_finished(self, work):
        episode = work.get_episode()
        quality = work.quality

        item = self._get_download_item(episode, quality)

        item.set_state(DownloadItemState.DONE)
        self.download_finished.emit(work)

    def _on_download_error(self, work, ex):
        episode = work.get_episode()
        quality = work.quality

        item = self._get_download_item(episode, quality)

        item.set_state(DownloadItemState.ERROR)
        item.set_error(ex)

    def _on_download_cancelled(self, work):
        episode = work.get_episode()
        quality = work.quality

        item = self._get_download_item(episode, quality)

        item.set_state(DownloadItemState.CANCELLED)
        self.download_cancelled.emit(work)

    def _on_timer_timeout(self):
        for func, args in self._delayed_update_calls:
            func(*args)

        self._delayed_update_calls = []

        self._signal_all_data_changed()

    def _signal_all_data_changed(self):
        index_start = self.createIndex(0, 0, None)
        last_row = len(self._download_list) - 1
        last_col = len(self._HEADER) - 1
        index_end = self.createIndex(last_row, last_col, None)

        self.dataChanged.emit(index_start, index_end)

    def exit(self):
        self._download_manager.exit()

    def index(self, row, column, parent):
        keys = list(self._download_list.keys())
        if row >= len(keys):
            return Qt.QModelIndex()
        key = keys[row]
        dl_item = self._download_list[key]
        work = dl_item.get_work()
        idx = self.createIndex(row, column, work)

        return idx

    def parent(self, child):
        return Qt.QModelIndex()

    def rowCount(self, parent):
        if not parent.isValid():
            return len(self._download_list)
        else:
            return 0

    def columnCount(self, parent):
        return len(QDownloadsTableModel._HEADER)

    @staticmethod
    def _format_size(size):
        if size < (1 << 10):
            s = '{} B'.format(size)
        elif size < (1 << 20):
            s = '{:.1f} kiB'.format(size / (1 << 10))
        elif size < (1 << 30):
            s = '{:.1f} MiB'.format(size / (1 << 20))
        else:
            s = '{:.1f} GiB'.format(size / (1 << 30))

        return s

    def data(self, index, role):
        col = index.column()
        if role == QtCore.Qt.DisplayRole:

            # I don't know why, calling index.internalPointer() seems to
            # segfault
            row = index.row()
            key = list(self._download_list.keys())[row]
            dl_item = self._download_list[key]
            dl_progress = dl_item.get_dl_progress()

            work = dl_item.get_work()
            episode = work.get_episode()
            quality = work.quality

            if col == 0:
                # Emission
                return episode.get_emission().get_title()
            elif col == 1:
                # Season/episode
                return episode.get_sae()
            elif col == 2:
                # Episode
                return episode.get_title()
            elif col == 3:
                # Filename
                filename = dl_item.get_filename()
                if filename is None:
                    filename = '?'

                return filename
            elif col == 4:
                # Segments
                done_segments = 0
                if dl_progress is not None:
                    done_segments = dl_progress.get_done_segments()
                total_segments = dl_item.get_total_segments()
                if total_segments is None:
                    total_segments = '?'

                return '{}/{}'.format(done_segments, total_segments)
            elif col == 5:
                # Downloaded bytes
                if dl_progress is None:
                    return 0

                done_bytes = dl_progress.get_done_bytes()
                dl = QDownloadsTableModel._format_size(done_bytes)

                return dl
            elif col == 6:
                # Estimated size
                estimated_size = dl_item.get_estimated_size()
                if estimated_size is None:
                    return '?'

                sz = QDownloadsTableModel._format_size(estimated_size)

                return sz
            elif col == 7:
                # Added date
                return dl_item.get_added_dt().strftime('%Y-%m-%d %H:%M:%S')
            elif col == 8:
                # Elapsed time
                total_seconds = dl_item.get_elapsed().seconds
                minutes = total_seconds // 60
                seconds = total_seconds - (minutes * 60)

                return '{}:{:02}'.format(minutes, seconds)
            elif col == 9:
                # Average download speed
                if dl_item.get_state() != DownloadItemState.RUNNING:
                    return '0 kiB/s'

                speed = dl_item.get_avg_download_speed()
                sz = QDownloadsTableModel._format_size(speed)

                return '{}/s'.format(sz)
            elif col == 10:
                # Progress bar
                return None
            elif col == 11:
                # Status
                handlers = QDownloadsTableModel._status_msg_handlers

                return handlers[dl_item.get_state()](dl_item)
            elif col == 12:
                return '{}x{}@{}kbps'.format(quality.xres,
                                             quality.yres,
                                             quality.bitrate // 1000)

    def headerData(self, col, ori, role):
        if ori == QtCore.Qt.Horizontal:
            if role == QtCore.Qt.DisplayRole:
                return QDownloadsTableModel._HEADER[col]

        return None



import os.path
from PyQt4 import QtCore
from PyQt4 import QtGui
from toutvqt import utils
from toutvqt.settings import SettingsKeys


class QTouTvPreferencesDialog(utils.QCommonDialog, utils.QtUiLoad):
    _UI_NAME = 'preferences_dialog'
    settings_accepted = QtCore.pyqtSignal(dict)

    def __init__(self, settings):
        super().__init__()

        self._setup_ui()
        self._setup_signals()
        self._setup_fields(settings)

    def _setup_fields(self, settings):
        dl_dir = settings.get_download_directory()
        proxy_url = settings.get_http_proxy()
        download_slots = settings.get_download_slots()
        always_max_quality = settings.get_always_max_quality()
        remove_finished = settings.get_remove_finished()

        self.http_proxy_value.setText(proxy_url)
        self.download_directory_value.setText(dl_dir)
        self.download_slots_value.setValue(download_slots)
        self.always_max_quality_check.setChecked(always_max_quality)
        self.remove_finished_check.setChecked(remove_finished)

    def _setup_signals(self):
        self.accepted.connect(self._send_settings_accepted)

    def _setup_ui(self):
        self._load_ui(QTouTvPreferencesDialog._UI_NAME)
        open_dl_dir_slot = self._open_download_directory_browser
        self.download_directory_browse.clicked.connect(open_dl_dir_slot)
        self._adjust_size()

    def _adjust_size(self):
        self.adjustSize()
        self.setFixedHeight(self.height())
        self.resize(600, self.height())

    def _open_download_directory_browser(self, checked):
        msg = 'Select download directory'
        dl_dir = QtGui.QFileDialog.getExistingDirectory(self, msg)
        if dl_dir.strip():
            self.download_directory_value.setText(os.path.abspath(dl_dir))

    def _send_settings_accepted(self):
        settings = {}

        dl_dir_value = self.download_directory_value.text().strip()
        proxy_url = self.http_proxy_value.text().strip()
        download_slots = self.download_slots_value.value()
        always_max_quality = self.always_max_quality_check.isChecked()
        remove_finished = self.remove_finished_check.isChecked()

        settings[SettingsKeys.NETWORK_HTTP_PROXY] = proxy_url
        settings[SettingsKeys.FILES_DOWNLOAD_DIR] = dl_dir_value
        settings[SettingsKeys.DL_DOWNLOAD_SLOTS] = download_slots
        settings[SettingsKeys.DL_ALWAYS_MAX_QUALITY] = always_max_quality
        settings[SettingsKeys.DL_REMOVE_FINISHED] = remove_finished

        self.settings_accepted.emit(settings)



#!/usr/bin/env python3
#
# Copyright (c) 2014, Philippe Proulx <eepp.ca>
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of the pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Philippe Proulx BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
from setuptools import setup
import toutv


# Make sure we run Python 3.3+ here
v = sys.version_info
if v.major < 3 or v.minor < 3:
    sys.stderr.write('Sorry, pytoutv needs Python 3.3+\n')
    sys.exit(1)

entry_points = {
    'console_scripts': [
        'toutv = toutvcli.app:run'
    ],
    'gui_scripts': [
        'qtoutv = toutvqt.app:run'
    ]
}
packages = [
    'toutv',
    'toutvcli',
    'toutvqt'
]
package_data = {
    'toutvqt': [
        'dat/ui/*.ui',
        'dat/icons/*.png'
    ]
}
install_requires = [
    'pycrypto>=2.0.0',
    'requests>=2.0.0',
    'setuptools>=3.0',
    # Workaround for bug on Travis with Python nightly
    'setuptools_scm',
]

setup_requires = [
    'pytest-runner',
]

tests_require = [
    'pytest',
]

setup(name='pytoutv',
      version=toutv.__version__,
      description='TOU.TV client library and user interfaces',
      author='Benjamin Vanheuverzwijn',
      author_email='bvanheu@gmail.com',
      url='https://github.com/bvanheu/pytoutv',
      keywords='TOUTV',
      license="BSD",
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      setup_requires=setup_requires,
      tests_require=tests_require,
      entry_points=entry_points,
      )



# Copyright (c) 2012, Benjamin Vanheuverzwijn <bvanheu@gmail.com>
# Copyright (c) 2014, Philippe Proulx <eepp.ca>
# All rights reserved.
#
# Thanks to Marc-Etienne M. Leveille
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#     * Redistributions of source code must retain the above copyright
#       notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright
#       notice, this list of conditions and the following disclaimer in the
#       documentation and/or other materials provided with the distribution.
#     * Neither the name of pytoutv nor the
#       names of its contributors may be used to endorse or promote products
#       derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
# ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL Benjamin Vanheuverzwijn OR Philippe Proulx
# BE LIABLE FOR ANY
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
# (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
# ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import errno
import struct
import logging
import requests
import functools
from Crypto.Cipher import AES
import toutv.config
import toutv.exceptions
import toutv.m3u8


class DownloadError(RuntimeError):

    def __init__(self, msg):
        self._msg = msg

    def __str__(self):
        return self._msg


class CancelledByUserError(DownloadError):

    def __init__(self):
        super().__init__('Download cancelled by user')


class FileExistsError(DownloadError):

    def __init__(self, path):
        super().__init__('File {} exists'.format(path))
        self._path = path

    @property
    def path(self):
        return self._path


class NoSpaceLeftError(DownloadError):

    def __init__(self):
        super().__init__('No space left on device')


class SegmentHandler:

    def initialize(self):
        """Called once before the Downloader tries to download any segment."""
        raise NotImplementedError()

    def has_segment(self, segindex):
        """Return whether the segment handler already has segment with index segindex.

        If the handler returns False, the Downloader will skip downloading
        this segment and call the segment_size method to determine the size
        of this segment.
        """
        raise NotImplementedError()

    def segment_size(self, segindex):
        """Return the size of segment with index segindex.

        This is called by the Downloader when the handler has indicated it
        already has a particular segment.
        """
        raise NotImplementedError()

    def on_segment(self, segindex, segment):
        """Called once for each downloaded segment.

        This method is not called for segments for which the has_segment
        returned True.
        """
        raise NotImplementedError()

    def finalize(self, num_segments):
        """Called once all the segments have been successfully downloaded."""
        raise NotImplementedError()


class FilesystemSegmentHandler(SegmentHandler):
    """SegmentHandler implementation which saves the episode on the filesystem."""

    def __init__(self,
                 episode,
                 bitrate,
                 output_dir,
                 filename,
                 overwrite=False):
        self._episode = episode
        self._bitrate = bitrate
        self._output_dir = output_dir
        self._overwrite = overwrite

        self._filename = filename
        self._output_path = os.path.join(self._output_dir, self._filename)

        self._logger = logging.getLogger(self.__class__.__name__)

    @property
    def filename(self):
        return self._filename

    @property
    def output_path(self):
        return self._output_path

    @property
    def output_dir(self):
        return self._output_dir

    def _get_segment_file_path(self, segindex):
        fmt = '.toutv-{}-{}-{}-{}.ts'
        segname = fmt.format(self._episode.get_emission().get_id(),
                             self._episode.get_id(),
                             self._bitrate,
                             segindex)

        return os.path.join(self._output_dir, segname)

    def _stitch_segment_files(self, num_segments):
        self._logger.debug('stitching {} segment files'.format(num_segments))
        part_output_path = self._output_path + '.part'

        with open(part_output_path, 'wb') as of:
            for segindex in range(num_segments):
                segpath = self._get_segment_file_path(segindex)

                if not os.path.isfile(segpath):
                    raise DownloadError('Cannot find segment file "{}"'.format(segpath))

                with open(segpath, 'rb') as segf:
                    self._logger.debug('concatenating segment file "{}"'.format(segpath))
                    of.write(segf.read())

        os.rename(part_output_path, self._output_path)

    def _remove_segment_file(self, segindex):
        segpath = self._get_segment_file_path(segindex)
        self._logger.debug('removing segment file "{}"'.format(segpath))

        try:
            os.remove(segpath)
        except Exception:
            # not the end of the world...
            self._logger.warn('cannot remove segment file "{}"'.format(segpath))

    def _remove_segment_files(self, num_segments):
        self._logger.debug('removing {} segment files'.format(num_segments))

        for segindex in range(num_segments):
            self._remove_segment_file(segindex)

    def initialize(self):
        self._logger.debug('episode: {}'.format(self._episode))
        self._logger.debug('bitrate: {}'.format(self._bitrate))
        self._logger.debug('output path: {}'.format(self._output_path))
        self._logger.debug('overwrite: {}'.format(self._overwrite))

        # Ensure the output directory exists.
        os.makedirs(self._output_dir, exist_ok=True)

        # prevent overwriting
        if not self._overwrite and os.path.exists(self._output_path):
            raise FileExistsError(self._output_path)

    def has_segment(self, segindex):
        segpath = self._get_segment_file_path(segindex)

        self._logger.debug('segment file path: "{}"'.format(segpath))

        return os.path.isfile(segpath)

    def segment_size(self, segindex):
        segpath = self._get_segment_file_path(segindex)
        statinfo = os.stat(segpath)
        return statinfo.st_size

    def on_segment(self, segindex, segment):
        segpath = self._get_segment_file_path(segindex)
        partpath = segpath + '.part'

        try:
            # completely write the part file first (could be interrupted)
            with open(partpath, 'wb') as f:
                self._logger.debug('writing partial segment file "{}"'.format(partpath))
                f.write(segment)

            # rename part file to segment file (should be atomic)
            os.rename(partpath, segpath)
        except OSError as e:
            if e.errno == errno.ENOSPC:
                raise NoSpaceLeftError()
            else:
                raise

    def finalize(self, num_segments):
        try:
            # stitch individual segment files as a complete file
            self._stitch_segment_files(num_segments)

            # remove segment files
            self._remove_segment_files(num_segments)
        except OSError as e:
            if e.errno == errno.ENOSPC:
                raise NoSpaceLeftError()
            else:
                raise


class SegmentProvider:

    def __init__(self):
        self.cancel = False

    def initialize(self):
        raise NotImplementedError()

    def num_segments(self):
        raise NotImplementedError()

    def download_segment(self, segindex, progress):
        raise NotImplementedError()

    def finalize(self):
        raise NotImplementedError()


class ToutvApiSegmentProvider(SegmentProvider):
    """Segment provider that fetches segments using the Tou.tv API"""

    _seg_aes_iv = struct.Struct('>IIII')

    def __init__(self, episode, bitrate, proxies=None, timeout=15):
        super().__init__()

        self._episode = episode
        self._bitrate = bitrate
        self._proxies = proxies
        self._timeout = timeout

        self._cookies = None
        self._video_playlist = None
        self._segments = None
        self._key = None

        self._logger = logging.getLogger(self.__class__.__name__)

    def _do_request(self, url, params=None, stream=False):
        self._logger.debug('HTTP GET request @ {}'.format(url))

        try:
            r = requests.get(url, params=params, headers=toutv.config.HEADERS,
                             proxies=self._proxies, cookies=self._cookies,
                             timeout=self._timeout, stream=stream)

            if r.status_code != 200:
                raise toutv.exceptions.UnexpectedHttpStatusCodeError(url,
                                                                     r.status_code)
        except requests.exceptions.Timeout:
            raise toutv.exceptions.RequestTimeoutError(url, self._timeout)
        except requests.exceptions.ConnectionError as e:
            raise toutv.exceptions.NetworkError() from e

        return r

    @staticmethod
    def _get_video_stream(playlist, bitrate):
        for stream in playlist.streams:
            if stream.bandwidth == bitrate:
                return stream

        raise DownloadError('Cannot find stream for bitrate {} bps'.format(bitrate))

    def _download_segment(self, segindex, progress):
        self._logger.debug('downloading segment {}'.format(segindex))

        encrypted_ts_segment = bytearray()
        chunks_count = 0
        num_bytes = 0

        # Obtain the URI to download this segment.
        segment = self._segments[segindex]
        request = self._do_request(segment.uri, stream=True)

        # Fetch by chunks of 8 kiB
        for chunk in request.iter_content(8192):
            if self.cancel:
                raise CancelledByUserError()

            encrypted_ts_segment += chunk
            num_bytes += len(chunk)

            # Every 32 chunks (256 kiB), we notify of our progress.
            if chunks_count % 32 == 0:
                progress(num_bytes)

            chunks_count += 1

        # We have the whole segment, decrypt it if needed.
        if self._key:
            aes_iv = self._seg_aes_iv.pack(0, 0, 0, segindex + 1)
            aes = AES.new(self._key, AES.MODE_CBC, aes_iv)
            ts_segment = aes.decrypt(bytes(encrypted_ts_segment))
        else:
            ts_segment = bytes(encrypted_ts_segment)

        return ts_segment

    def _download_segment_with_retry(self, segindex, progress, num_tries=3):
        for i in range(num_tries):
            try:
                return self._download_segment(segindex, progress)
            except toutv.exceptions.NetworkError:
                # If it was our last retry, give up and propagate the exception.
                if i + 1 == num_tries:
                    raise

    def initialize(self):
        self._logger.debug('episode: {}'.format(self._episode))
        self._logger.debug('bitrate: {}'.format(self._bitrate))
        self._logger.debug('timeout: {}'.format(self._timeout))

        playlist, cookies = self._episode.get_playlist_cookies()
        self._cookies = cookies

        # select appropriate stream for required bitrate
        stream = self._get_video_stream(playlist, self._bitrate)

        # get video playlist
        m3u8_file = self._do_request(stream.uri).text
        self._video_playlist = toutv.m3u8.parse(m3u8_file,
                                                os.path.dirname(stream.uri))
        self._segments = self._video_playlist.segments
        self._logger.debug('parsed M3U8 file: {} total segments'.format(self.num_segments()))

        # get decryption key
        if self._segments[0].key:
            uri = self._segments[0].key.uri
            self._key = self._do_request(uri).content
            self._logger.debug('decryption key: {}'.format(self._key))
        else:
            self._logger.debug('no decryption key found')

    def num_segments(self):
        return len(self._segments)

    def download_segment(self, segindex, progress):
        return self._download_segment_with_retry(segindex, progress)

    def finalize(self):
        pass


class Downloader:

    def __init__(self,
                 seg_provider,
                 seg_handler,
                 on_progress_update=None,
                 on_dl_start=None):
        self._seg_provider = seg_provider
        self._seg_handler = seg_handler

        self._on_progress_update = on_progress_update
        self._on_dl_start = on_dl_start

        self._do_cancel = False
        self._logger = logging.getLogger(self.__class__.__name__)

    def cancel(self):
        self._logger.info('cancelling download')
        self._seg_provider.cancel = True
        self._do_cancel = True

    def _notify_dl_start(self, num_segments):
        if self._on_dl_start:
            self._on_dl_start(num_segments)

    def _notify_progress_update(self, num_completed_segments, num_bytes,
                                num_bytes_partial_segment):
        if self._on_progress_update:
            self._on_progress_update(num_completed_segments, num_bytes,
                                     num_bytes_partial_segment)

    def download(self):
        self._logger.debug('starting download')

        self._seg_handler.initialize()
        self._seg_provider.initialize()

        # Get the number of segments.
        num_segments = self._seg_provider.num_segments()

        # Notify of the download start.
        self._notify_dl_start(num_segments)

        # Do an initial progress update before we begin.
        self._notify_progress_update(0, 0, 0)

        # Number of bytes in the completely downloaded segments.
        done_segment_bytes = 0

        try:
            for segindex in range(num_segments):

                if self._do_cancel:
                    raise CancelledByUserError()

                if self._seg_handler.has_segment(segindex):
                    self._logger.debug('segment handler already has segment; skipping')
                    done_segment_bytes += self._seg_handler.segment_size(segindex)
                    continue

                # Function called by the segment provider to notify of progress
                # during the fetching of a segment.
                progress = functools.partial(self._notify_progress_update,
                                             segindex, done_segment_bytes)

                # Get the segment.
                segment = self._seg_provider.download_segment(segindex, progress)

                # Update running sum of bytes.
                done_segment_bytes += len(segment)

                # Notify of progress.
                self._notify_progress_update(segindex + 1, done_segment_bytes, 0)

                # Do something with the segment.
                self._seg_handler.on_segment(segindex, segment)

            # All the segments were fetched.
            self._seg_provider.finalize()
            self._seg_handler.finalize(num_segments)
        except DownloadError as e:
            # If the exception is already a DownloadError, just propagate it...
            raise e
        except Exception as e:
            # ... otherwise, throw a DownloadError from the original exception.
            tmpl = 'Download error: {}'
            raise DownloadError(tmpl.format(e)) from e



import logging
from PyQt4 import Qt
from PyQt4 import QtGui
from PyQt4 import QtCore
import webbrowser
from toutvqt import utils


class QInfosFrame(Qt.QFrame):
    select_download = QtCore.pyqtSignal(list)

    def __init__(self, client):
        super().__init__()

        self._client = client

        self._setup_thumb_fetching()
        self._setup_ui()
        self.show_infos_none()

    def _swap_infos_widget(self, widget):
        for swappable_widget in self._swappable_widgets:
            if widget is not swappable_widget:
                swappable_widget.hide()
        widget.show()

    def exit(self):
        logging.debug('Joining thumb fetcher thread')
        self._fetch_thumb_thread.quit()
        self._fetch_thumb_thread.wait()

    def show_infos_none(self):
        logging.debug('Showing none label')
        self._swap_infos_widget(self.none_label)

    def show_emission(self, emission):
        logging.debug('Showing emission infos')
        self.emission_widget.set_emission(emission)
        self._swap_infos_widget(self.emission_widget)

    def show_season(self, emission, season_number, episodes):
        logging.debug('Showing season infos')
        self.season_widget.set_infos(emission, season_number, episodes)
        self._swap_infos_widget(self.season_widget)

    def show_episode(self, episode):
        logging.debug('Showing episode infos')
        self.episode_widget.set_episode(episode)
        self._swap_infos_widget(self.episode_widget)

    def _setup_none_label(self):
        self.none_label = Qt.QLabel()
        self.none_label.setText('Please select an item in the list above')
        font = Qt.QFont()
        font.setItalic(True)
        self.none_label.setFont(font)

    def _setup_infos_widget(self):
        self._setup_none_label()
        self.emission_widget = _QEmissionInfosWidget(self._thumb_fetcher,
                                                     self._client)
        self.emission_widget.select_download.connect(self.select_download)
        self.season_widget = _QSeasonInfosWidget()
        self.season_widget.select_download.connect(self.select_download)
        self.episode_widget = _QEpisodeInfosWidget(self._thumb_fetcher)
        self.episode_widget.select_download.connect(self.select_download)

        self._swappable_widgets = [
            self.emission_widget,
            self.season_widget,
            self.episode_widget,
            self.none_label,
        ]

        for widget in self._swappable_widgets:
            widget.hide()
            self.layout().addWidget(widget)

    def _setup_ui(self):
        self.setLayout(Qt.QVBoxLayout())
        self.setFrameShape(Qt.QFrame.Box)
        self.setFrameShadow(Qt.QFrame.Sunken)
        self._setup_infos_widget()
        self.setSizePolicy(QtGui.QSizePolicy.Expanding,
                           QtGui.QSizePolicy.Maximum)

    def _setup_thumb_fetching(self):
        self._fetch_thumb_thread = Qt.QThread()
        self._fetch_thumb_thread.start()

        self._thumb_fetcher = _QThumbFetcher()
        self._thumb_fetcher.moveToThread(self._fetch_thumb_thread)


class _QThumbFetcher(Qt.QObject):
    fetch_done = QtCore.pyqtSignal(object)

    def __init__(self):
        super().__init__()

        self._last = None

    def set_last(self, bo):
        self._last = bo

    def fetch_thumb(self, bo):
        if bo is not self._last:
            tmpl = 'Skipping thumbnail fetching of "{}"'
            logging.debug(tmpl.format(bo.get_title()))
            return

        tmpl = 'Fetching thumbnail for episode "{}"'
        logging.debug(tmpl.format(bo.get_title()))
        bo.get_medium_thumb_data()
        self.fetch_done.emit(bo)


class _QInfosWidget(Qt.QWidget, utils.QtUiLoad):
    _fetch_thumb_required = QtCore.pyqtSignal(object)
    select_download = QtCore.pyqtSignal(object)

    def __init__(self, thumb_fetcher):
        super().__init__()

        self._thumb_fetcher = thumb_fetcher
        self._bo = None
        self._url = None

    def _setup_ui(self, ui_name):
        self._load_ui(ui_name)
        self.goto_toutv_btn.clicked.connect(self._on_goto_toutv_btn_clicked)
        self.dl_btn.clicked.connect(self._on_dl_btn_clicked)

    def _set_toutv_url(self, url):
        self._url = url
        if url is None:
            self.goto_toutv_btn.hide()
        else:
            self.goto_toutv_btn.show()

    def _on_goto_toutv_btn_clicked(self):
        if self._url is not None:
            logging.debug('Going to TOU.TV @ "{}"'.format(self._url))
            webbrowser.open(self._url)

    def _on_dl_btn_clicked(self):
        logging.debug('Download button clicked')
        pass

    def _setup_thumb_fetching(self):
        # Setup signal connections with thumb fetcher
        self._fetch_thumb_required.connect(self._thumb_fetcher.fetch_thumb)
        self._thumb_fetcher.fetch_done.connect(self._thumb_fetched)

    def _set_no_thumb(self):
        self.thumb_value_label.setPixmap(Qt.QPixmap())

    def _set_thumb(self):
        jpeg_data = self._bo.get_medium_thumb_data()
        if jpeg_data is None:
            self._set_no_thumb()

        pixmap = Qt.QPixmap()
        ret = pixmap.loadFromData(jpeg_data, 'JPEG')
        if not ret:
            self._set_no_thumb()
            return

        smooth_transform = QtCore.Qt.SmoothTransformation
        width = self.thumb_value_label.width()
        scaled_pixmap = pixmap.scaledToWidth(width, smooth_transform)
        self.thumb_value_label.setPixmap(scaled_pixmap)

    def _try_set_thumb(self):
        if self._bo.has_medium_thumb_data():
            self._set_thumb()
        else:
            self._set_no_thumb()
            self._thumb_fetcher.set_last(self._bo)
            self._fetch_thumb_required.emit(self._bo)

    def _thumb_fetched(self, bo):
        if bo is not self._bo:
            # Not us, or too late. Ignore; next time will be faster anyway.
            return

        self._set_thumb()


class _QEmissionCommonInfosWidget:
    def _set_removal_date(self):
        removal_date = self._bo.get_removal_date()
        if removal_date is None:
            removal_date = '-'
        else:
            removal_date = str(removal_date)
        self.removal_date_value_label.setText(removal_date)

    def _set_genre(self):
        genre = self._bo.get_genre()
        if genre is None:
            genre = '-'
        else:
            genre = genre.get_title()
        self.genre_value_label.setText(genre)

    def _set_network(self):
        network = self._bo.get_network()
        if network is None:
            network = '-'
        self.network_value_label.setText(network)

    def _set_country(self):
        country = self._bo.get_country()
        if country is None:
            country = '-'
        self.country_value_label.setText(country)

    def _set_common_infos(self):
        self._set_removal_date()
        self._set_genre()
        self._set_network()
        self._set_country()


class _QEmissionInfosWidget(_QInfosWidget, _QEmissionCommonInfosWidget):
    _UI_NAME = 'emission_infos_widget'
    _fetch_thumb_required = QtCore.pyqtSignal(object)

    def __init__(self, thumb_fetcher, client):
        super().__init__(thumb_fetcher)

        self._client = client

        self._setup_ui(_QEmissionInfosWidget._UI_NAME)
        self._setup_thumb_fetching()

    def _setup_ui(self, ui_name):
        super()._setup_ui(ui_name)
        width = self.thumb_value_label.width()
        min_height = round(width * 9 / 16) + 1
        self.thumb_value_label.setMinimumHeight(min_height)

    def _set_title(self):
        self.title_value_label.setText(self._bo.get_title())

    def _set_description(self):
        description = self._bo.get_description()
        if description is None:
            description = ''
        self.description_value_label.setText(description)

    def set_emission(self, emission):
        self._bo = emission

        self._set_title()
        self._set_description()
        self._set_common_infos()
        self._set_toutv_url(emission.get_url())
        self._try_set_thumb()

    def _on_dl_btn_clicked(self):
        episodes = self._client.get_emission_episodes(self._bo)
        self.select_download.emit(list(episodes))


class _QSeasonInfosWidget(_QInfosWidget, _QEmissionCommonInfosWidget):
    _UI_NAME = 'season_infos_widget'

    def __init__(self):
        super().__init__(None)

        self._setup_ui(_QSeasonInfosWidget._UI_NAME)

    def _set_season_number(self):
        self.season_number_value_label.setText(str(self._season_number))

    def _set_number_episodes(self):
        self.number_episodes_value_label.setText(str(len(self._episodes)))

    def set_infos(self, emission, season_number, episodes):
        self._bo = emission
        self._season_number = season_number
        self._episodes = [e.bo for e in episodes]

        self._set_season_number()
        self._set_number_episodes()
        self._set_common_infos()
        self._set_toutv_url(emission.get_url())

    def _on_dl_btn_clicked(self):
        self.select_download.emit(self._episodes)


class _QEpisodeInfosWidget(_QInfosWidget):
    _UI_NAME = 'episode_infos_widget'

    def __init__(self, thumb_fetcher):
        super().__init__(thumb_fetcher)

        self._setup_ui(_QEpisodeInfosWidget._UI_NAME)
        self._setup_thumb_fetching()

    def _setup_ui(self, ui_name):
        super()._setup_ui(ui_name)
        width = self.thumb_value_label.width()
        min_height = round(width * 9 / 16) + 1
        self.thumb_value_label.setMinimumHeight(min_height)

    def _set_description(self):
        description = self._bo.get_description()
        if description is None:
            description = '-'
        self.description_value_label.setText(description)

    def _set_air_date(self):
        air_date = self._bo.get_air_date()
        if air_date is None:
            air_date = '-'
        self.air_date_value_label.setText(str(air_date))

    def _set_length(self):
        minutes, seconds = self._bo.get_length()
        length = '{}:{:02}'.format(minutes, seconds)
        self.length_value_label.setText(length)

    def _set_sae(self):
        sae = self._bo.get_sae()
        if sae is None:
            sae = '-'
        self.sae_value_label.setText(sae)

    def _set_director(self):
        director = self._bo.get_director()
        if director is None:
            director = '-'
        self.director_value_label.setText(director)

    def _set_author(self):
        author = self._bo.get_author()
        if author is None:
            author = '-'
        self.author_value_label.setText(author)

    def _set_titles(self):
        emission = self._bo.get_emission()
        self.title_value_label.setText(self._bo.get_title())
        self.emission_title_value_label.setText(emission.get_title())

    def set_episode(self, episode):
        self._bo = episode

        self._set_titles()
        self._set_author()
        self._set_director()
        self._set_description()
        self._set_air_date()
        self._set_length()
        self._set_sae()
        self._try_set_thumb()
        url = '{}?autoplay=true'.format(episode.get_url())
        self._set_toutv_url(url)

    def _on_dl_btn_clicked(self):
        self.select_download.emit([self._bo])



from PyQt4 import Qt, QtCore
from toutvqt.downloads_itemdelegate import QDlItemDelegate
from toutvqt.downloads_tablemodel import DownloadItemState
from toutvqt import utils


class QDownloadsTableView(Qt.QTreeView):
    def __init__(self, model):
        super().__init__()

        self.setRootIsDecorated(False)
        self.setItemDelegate(QDlItemDelegate(model))
        self._setup(model)

    def _build_context_menu(self):
        self._context_menu = Qt.QMenu(parent=self)

        # Actions
        self._remove_item_action = self._context_menu.addAction('Remove item')
        self._cancel_action = self._context_menu.addAction('Cancel')
        self._open_action = self._context_menu.addAction('Open')
        self._open_dir_action = self._context_menu.addAction('Open directory')

        # Icons
        self._remove_item_action.setIcon(utils.get_qicon('remove_item_action'))
        self._open_action.setIcon(utils.get_qicon('open_action'))
        self._open_dir_action.setIcon(utils.get_qicon('open_dir_action'))
        self._cancel_action.setIcon(utils.get_qicon('cancel_action'))

        self._visible_context_menu_actions = {
            DownloadItemState.QUEUED: [
                self._cancel_action,
                self._open_dir_action,
            ],
            DownloadItemState.RUNNING: [
                self._cancel_action,
                self._open_action,
                self._open_dir_action,
            ],
            DownloadItemState.PAUSED: [
                self._cancel_action,
                self._open_action,
                self._open_dir_action,
            ],
            DownloadItemState.CANCELLED: [
                self._remove_item_action,
                self._open_dir_action,
            ],
            DownloadItemState.ERROR: [
                self._remove_item_action,
                self._open_dir_action,
            ],
            DownloadItemState.DONE: [
                self._remove_item_action,
                self._open_action,
                self._open_dir_action,
            ],
        }

    def _setup_context_menu(self):
        self._build_context_menu()
        self.setContextMenuPolicy(QtCore.Qt.CustomContextMenu)
        self.customContextMenuRequested.connect(self._on_context_menu)

    def _setup(self, model):
        self.setModel(model)
        self._setup_context_menu()

    def _arrange_context_menu(self, dl_item_state):
        self._remove_item_action.setVisible(False)
        self._cancel_action.setVisible(False)
        self._open_action.setVisible(False)
        self._open_dir_action.setVisible(False)

        for action in self._visible_context_menu_actions[dl_item_state]:
            action.setVisible(True)

    def _on_context_menu(self, pos):
        index = self.indexAt(pos)
        if not index.isValid():
            return

        dl_item = self.model().get_download_item_at_row(index.row())

        self._arrange_context_menu(dl_item.get_state())
        action = self._context_menu.exec(Qt.QCursor.pos())

        if action is self._open_action:
            output_dir = dl_item.get_work().get_output_dir()
            filename = dl_item.get_filename()
            url = Qt.QUrl('file://{}/{}'.format(output_dir, filename))
            Qt.QDesktopServices.openUrl(url)
        elif action is self._open_dir_action:
            output_dir = dl_item.get_work().get_output_dir()
            url = Qt.QUrl('file://{}'.format(output_dir))
            Qt.QDesktopServices.openUrl(url)
        elif action is self._cancel_action:
            self.model().cancel_download_at_row(index.row())
        elif action is self._remove_item_action:
            self.model().remove_item_at_row(index.row())

    def set_default_columns_widths(self):
        pass



from PyQt4 import Qt
from PyQt4 import QtCore


class QDlItemDelegate(Qt.QItemDelegate):

    def __init__(self, model):
        super().__init__(model)
        self._model = model

    @staticmethod
    def _get_progress_bar(option, percent):
        p = Qt.QStyleOptionProgressBarV2()

        p.state = Qt.QStyle.State_Enabled
        p.direction = Qt.QApplication.layoutDirection()
        p.rect = option.rect
        p.fontMetrics = Qt.QApplication.fontMetrics()
        p.minimum = 0
        p.maximum = 100
        p.textAlignment = QtCore.Qt.AlignCenter
        p.textVisible = True
        p.progress = percent
        p.text = '{}%'.format(percent)

        return p

    def paint(self, painter, option, index):
        # Mostly taken from:
        # http://qt-project.org/doc/qt-4.8/network-torrent-mainwindow-cpp.html

        if index.column() != self._model.get_progress_col():
            super().paint(painter, option, index)
            return

        dl_item = self._model.get_download_item_at_row(index.row())
        percent = dl_item.get_progress_percent()

        progress_bar = QDlItemDelegate._get_progress_bar(option, percent)

        style = Qt.QApplication.style()
        style.drawControl(Qt.QStyle.CE_ProgressBar, progress_bar, painter)
